package org.eclipse.jdt.internal.compiler.lookup;

import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;
import java.util.Iterator;
import org.eclipse.jdt.internal.compiler.problem.AbortCompilation;
import java.util.Map;
import java.util.HashMap;
import org.eclipse.jdt.internal.compiler.ast.QualifiedAllocationExpression;
import org.eclipse.jdt.internal.compiler.ast.TypeParameter;
import org.eclipse.jdt.internal.compiler.env.AccessRestriction;
import org.eclipse.jdt.internal.compiler.ast.FieldDeclaration;
import org.eclipse.jdt.internal.compiler.util.HashtableOfObject;
import org.eclipse.jdt.internal.compiler.ast.AbstractMethodDeclaration;
import org.eclipse.jdt.core.compiler.CharOperation;
import java.util.ArrayList;
import org.eclipse.jdt.internal.compiler.ast.TypeReference;
import org.eclipse.jdt.internal.compiler.ast.TypeDeclaration;

public class ClassScope extends Scope
{
    public TypeDeclaration referenceContext;
    public TypeReference superTypeReference;
    ArrayList<Object> deferredBoundChecks;
    
    public ClassScope(Scope parent, TypeDeclaration context) {
        super(3, parent);
        this.referenceContext = context;
        this.deferredBoundChecks = null;
    }
    
    void buildAnonymousTypeBinding(SourceTypeBinding enclosingType, ReferenceBinding supertype) {
        LocalTypeBinding anonymousType = this.buildLocalType(enclosingType, enclosingType.fPackage);
        anonymousType.modifiers |= 0x8000000;
        int inheritedBits = supertype.typeBits;
        if ((inheritedBits & 0x4) != 0x0) {
            AbstractMethodDeclaration[] methods = this.referenceContext.methods;
            if (methods != null) {
                int i = 0;
                while (i < methods.length) {
                    if (CharOperation.equals(TypeConstants.CLOSE, methods[i].selector) && methods[i].arguments == null) {
                        inheritedBits &= 0x13;
                        break;
                    }
                    ++i;
                }
            }
        }
        anonymousType.typeBits |= inheritedBits;
        if (supertype.isInterface()) {
            anonymousType.setSuperClass(this.getJavaLangObject());
            anonymousType.setSuperInterfaces(new ReferenceBinding[] { supertype });
            TypeReference typeReference = this.referenceContext.allocation.type;
            if (typeReference != null) {
                this.referenceContext.superInterfaces = new TypeReference[] { typeReference };
                if ((supertype.tagBits & 0x40000000L) != 0x0L) {
                    this.problemReporter().superTypeCannotUseWildcard(anonymousType, typeReference, supertype);
                    anonymousType.tagBits |= 0x20000L;
                    anonymousType.setSuperInterfaces(Binding.NO_SUPERINTERFACES);
                }
            }
        }
        else {
            anonymousType.setSuperClass(supertype);
            anonymousType.setSuperInterfaces(Binding.NO_SUPERINTERFACES);
            TypeReference typeReference = this.referenceContext.allocation.type;
            if (typeReference != null) {
                this.referenceContext.superclass = typeReference;
                if (supertype.erasure().id == 41) {
                    this.problemReporter().cannotExtendEnum(anonymousType, typeReference, supertype);
                    anonymousType.tagBits |= 0x20000L;
                    anonymousType.setSuperClass(this.getJavaLangObject());
                }
                else if (supertype.isFinal()) {
                    this.problemReporter().anonymousClassCannotExtendFinalClass(typeReference, supertype);
                    anonymousType.tagBits |= 0x20000L;
                    anonymousType.setSuperClass(this.getJavaLangObject());
                }
                else if ((supertype.tagBits & 0x40000000L) != 0x0L) {
                    this.problemReporter().superTypeCannotUseWildcard(anonymousType, typeReference, supertype);
                    anonymousType.tagBits |= 0x20000L;
                    anonymousType.setSuperClass(this.getJavaLangObject());
                }
            }
        }
        this.connectMemberTypes();
        this.buildFieldsAndMethods();
        anonymousType.faultInTypesForFieldsAndMethods();
        anonymousType.verifyMethods(this.environment().methodVerifier());
    }
    
    void buildFields() {
        SourceTypeBinding sourceType = this.referenceContext.binding;
        if (sourceType.areFieldsInitialized()) {
            return;
        }
        if (this.referenceContext.fields == null) {
            sourceType.setFields(Binding.NO_FIELDS);
            return;
        }
        FieldDeclaration[] fields = this.referenceContext.fields;
        int size = fields.length;
        int count = 0;
        for (int i = 0; i < size; ++i) {
            switch (fields[i].getKind()) {
                case 1:
                case 3: {
                    ++count;
                    break;
                }
            }
        }
        FieldBinding[] fieldBindings = new FieldBinding[count];
        HashtableOfObject knownFieldNames = new HashtableOfObject(count);
        count = 0;
        for (int j = 0; j < size; ++j) {
            FieldDeclaration field = fields[j];
            if (field.getKind() != 2) {
                FieldBinding fieldBinding = new FieldBinding(field, null, field.modifiers | 0x2000000, sourceType);
                fieldBinding.id = count;
                this.checkAndSetModifiersForField(fieldBinding, field);
                if (knownFieldNames.containsKey(field.name)) {
                    FieldBinding previousBinding = (FieldBinding)knownFieldNames.get(field.name);
                    if (previousBinding != null) {
                        int f = 0;
                        while (f < j) {
                            FieldDeclaration previousField = fields[f];
                            if (previousField.binding == previousBinding) {
                                this.problemReporter().duplicateFieldInType(sourceType, previousField);
                                break;
                            }
                            ++f;
                        }
                    }
                    knownFieldNames.put(field.name, null);
                    this.problemReporter().duplicateFieldInType(sourceType, field);
                    field.binding = null;
                }
                else {
                    knownFieldNames.put(field.name, fieldBinding);
                    fieldBindings[count++] = fieldBinding;
                }
            }
        }
        if (count != fieldBindings.length) {
            System.arraycopy(fieldBindings, 0, fieldBindings = new FieldBinding[count], 0, count);
        }
        sourceType.tagBits &= 0xFFFFFFFFFFFFCFFFL;
        sourceType.setFields(fieldBindings);
        return;
    }
    
    void buildFieldsAndMethods() {
        this.buildFields();
        this.buildMethods();
        SourceTypeBinding sourceType = this.referenceContext.binding;
        if (!sourceType.isPrivate() && sourceType.superclass instanceof SourceTypeBinding && sourceType.superclass.isPrivate()) {
            ((SourceTypeBinding)sourceType.superclass).tagIndirectlyAccessibleMembers();
        }
        if (sourceType.isMemberType() && !sourceType.isLocalType()) {
            ((MemberTypeBinding)sourceType).checkSyntheticArgsAndFields();
        }
        ReferenceBinding[] memberTypes = sourceType.memberTypes;
        for (int i = 0, length = memberTypes.length; i < length; ++i) {
            ((SourceTypeBinding)memberTypes[i]).scope.buildFieldsAndMethods();
        }
    }
    
    private LocalTypeBinding buildLocalType(SourceTypeBinding enclosingType, PackageBinding packageBinding) {
        this.referenceContext.scope = this;
        this.referenceContext.staticInitializerScope = new MethodScope(this, this.referenceContext, true);
        this.referenceContext.initializerScope = new MethodScope(this, this.referenceContext, false);
        LocalTypeBinding localType = new LocalTypeBinding(this, enclosingType, this.innermostSwitchCase());
        this.referenceContext.binding = localType;
        this.checkAndSetModifiers();
        this.buildTypeVariables();
        ReferenceBinding[] memberTypeBindings = Binding.NO_MEMBER_TYPES;
        if (this.referenceContext.memberTypes != null) {
            int size = this.referenceContext.memberTypes.length;
            memberTypeBindings = new ReferenceBinding[size];
            int count = 0;
        Label_0315:
            for (int i = 0; i < size; ++i) {
                TypeDeclaration memberContext = this.referenceContext.memberTypes[i];
                switch (TypeDeclaration.kind(memberContext.modifiers)) {
                    case 2:
                    case 4: {
                        this.problemReporter().illegalLocalTypeDeclaration(memberContext);
                        break;
                    }
                    default: {
                        ReferenceBinding type = localType;
                        while (!CharOperation.equals(type.sourceName, memberContext.name)) {
                            type = type.enclosingType();
                            if (type != null) {
                                continue;
                            }
                            int j = 0;
                            while (j < i) {
                                if (CharOperation.equals(this.referenceContext.memberTypes[j].name, memberContext.name)) {
                                    this.problemReporter().duplicateNestedType(memberContext);
                                    continue Label_0315;
                                }
                                ++j;
                            }
                            ClassScope memberScope = new ClassScope(this, this.referenceContext.memberTypes[i]);
                            LocalTypeBinding memberBinding = memberScope.buildLocalType(localType, packageBinding);
                            memberBinding.setAsMemberType();
                            memberTypeBindings[count++] = memberBinding;
                            continue Label_0315;
                        }
                        this.problemReporter().typeCollidesWithEnclosingType(memberContext);
                        break;
                    }
                }
            }
            if (count != size) {
                System.arraycopy(memberTypeBindings, 0, memberTypeBindings = new ReferenceBinding[count], 0, count);
            }
        }
        localType.setMemberTypes(memberTypeBindings);
        return localType;
    }
    
    void buildLocalTypeBinding(SourceTypeBinding enclosingType) {
        LocalTypeBinding localType = this.buildLocalType(enclosingType, enclosingType.fPackage);
        this.connectTypeHierarchy();
        if (this.compilerOptions().sourceLevel >= 3211264L) {
            this.checkParameterizedTypeBounds();
            this.checkParameterizedSuperTypeCollisions();
        }
        this.buildFieldsAndMethods();
        localType.faultInTypesForFieldsAndMethods();
        this.referenceContext.binding.verifyMethods(this.environment().methodVerifier());
    }
    
    private void buildMemberTypes(AccessRestriction accessRestriction) {
        SourceTypeBinding sourceType = this.referenceContext.binding;
        ReferenceBinding[] memberTypeBindings = Binding.NO_MEMBER_TYPES;
        if (this.referenceContext.memberTypes != null) {
            int length = this.referenceContext.memberTypes.length;
            memberTypeBindings = new ReferenceBinding[length];
            int count = 0;
            int i = 0;
            while (i < length) {
                TypeDeclaration memberContext = this.referenceContext.memberTypes[i];
                if (this.environment().isProcessingAnnotations && this.environment().isMissingType(memberContext.name)) {
                    throw new SourceTypeCollisionException();
                }
                Label_0286: {
                    switch (TypeDeclaration.kind(memberContext.modifiers)) {
                        case 2:
                        case 4: {
                            if (sourceType.isNestedType() && sourceType.isClass() && !sourceType.isStatic()) {
                                this.problemReporter().illegalLocalTypeDeclaration(memberContext);
                                break Label_0286;
                            }
                            break;
                        }
                    }
                    ReferenceBinding type = sourceType;
                    while (!CharOperation.equals(type.sourceName, memberContext.name)) {
                        type = type.enclosingType();
                        if (type != null) {
                            continue;
                        }
                        int j = 0;
                        while (j < i) {
                            if (CharOperation.equals(this.referenceContext.memberTypes[j].name, memberContext.name)) {
                                this.problemReporter().duplicateNestedType(memberContext);
                                break Label_0286;
                            }
                            ++j;
                        }
                        ClassScope memberScope = new ClassScope(this, memberContext);
                        memberTypeBindings[count++] = memberScope.buildType(sourceType, sourceType.fPackage, accessRestriction);
                        break Label_0286;
                    }
                    this.problemReporter().typeCollidesWithEnclosingType(memberContext);
                }
                ++i;
            }
            if (count != length) {
                System.arraycopy(memberTypeBindings, 0, memberTypeBindings = new ReferenceBinding[count], 0, count);
            }
        }
        sourceType.setMemberTypes(memberTypeBindings);
    }
    
    void buildMethods() {
        SourceTypeBinding sourceType = this.referenceContext.binding;
        if (sourceType.areMethodsInitialized()) {
            return;
        }
        boolean isEnum = TypeDeclaration.kind(this.referenceContext.modifiers) == 3;
        if (this.referenceContext.methods == null && !isEnum) {
            this.referenceContext.binding.setMethods(Binding.NO_METHODS);
            return;
        }
        AbstractMethodDeclaration[] methods = this.referenceContext.methods;
        int size = (methods == null) ? 0 : methods.length;
        int clinitIndex = -1;
        int i = 0;
        while (i < size) {
            if (methods[i].isClinit()) {
                clinitIndex = i;
                break;
            }
            ++i;
        }
        int count = isEnum ? 2 : 0;
        MethodBinding[] methodBindings = new MethodBinding[((clinitIndex == -1) ? size : (size - 1)) + count];
        if (isEnum) {
            methodBindings[0] = sourceType.addSyntheticEnumMethod(TypeConstants.VALUES);
            methodBindings[1] = sourceType.addSyntheticEnumMethod(TypeConstants.VALUEOF);
        }
        boolean hasNativeMethods = false;
        if (sourceType.isAbstract()) {
            for (int k = 0; k < size; ++k) {
                if (k != clinitIndex) {
                    MethodScope scope2 = new MethodScope(this, methods[k], false);
                    MethodBinding methodBinding2 = scope2.createMethod(methods[k]);
                    if (methodBinding2 != null) {
                        methodBindings[count++] = methodBinding2;
                        hasNativeMethods = (hasNativeMethods || methodBinding2.isNative());
                    }
                }
            }
        }
        else {
            boolean hasAbstractMethods = false;
            for (int j = 0; j < size; ++j) {
                if (j != clinitIndex) {
                    MethodScope scope = new MethodScope(this, methods[j], false);
                    MethodBinding methodBinding = scope.createMethod(methods[j]);
                    if (methodBinding != null) {
                        methodBindings[count++] = methodBinding;
                        hasAbstractMethods = (hasAbstractMethods || methodBinding.isAbstract());
                        hasNativeMethods = (hasNativeMethods || methodBinding.isNative());
                    }
                }
            }
            if (hasAbstractMethods) {
                this.problemReporter().abstractMethodInConcreteClass(sourceType);
            }
        }
        if (count != methodBindings.length) {
            System.arraycopy(methodBindings, 0, methodBindings = new MethodBinding[count], 0, count);
        }
        sourceType.tagBits &= 0xFFFFFFFFFFFF3FFFL;
        sourceType.setMethods(methodBindings);
        if (hasNativeMethods) {
            for (int k = 0; k < methodBindings.length; ++k) {
                MethodBinding methodBinding3 = methodBindings[k];
                methodBinding3.modifiers |= 0x8000000;
            }
            FieldBinding[] fields = sourceType.unResolvedFields();
            for (int j = 0; j < fields.length; ++j) {
                FieldBinding fieldBinding = fields[j];
                fieldBinding.modifiers |= 0x8000000;
            }
        }
        if (isEnum && this.compilerOptions().isAnnotationBasedNullAnalysisEnabled) {
            LookupEnvironment environment = this.environment();
            ((SyntheticMethodBinding)methodBindings[0]).markNonNull(environment);
            ((SyntheticMethodBinding)methodBindings[1]).markNonNull(environment);
        }
        return;
    }
    
    SourceTypeBinding buildType(SourceTypeBinding enclosingType, PackageBinding packageBinding, AccessRestriction accessRestriction) {
        this.referenceContext.scope = this;
        this.referenceContext.staticInitializerScope = new MethodScope(this, this.referenceContext, true);
        this.referenceContext.initializerScope = new MethodScope(this, this.referenceContext, false);
        if (enclosingType == null) {
            char[][] className = CharOperation.arrayConcat(packageBinding.compoundName, this.referenceContext.name);
            this.referenceContext.binding = new SourceTypeBinding(className, packageBinding, this);
        }
        else {
            char[][] className = CharOperation.deepCopy(enclosingType.compoundName);
            className[className.length - 1] = CharOperation.concat(className[className.length - 1], this.referenceContext.name, '$');
            ReferenceBinding existingType = packageBinding.getType0(className[className.length - 1]);
            if (existingType != null && !(existingType instanceof UnresolvedReferenceBinding)) {
                this.parent.problemReporter().duplicateNestedType(this.referenceContext);
            }
            this.referenceContext.binding = new MemberTypeBinding(className, this, enclosingType);
        }
        SourceTypeBinding sourceType = this.referenceContext.binding;
        this.environment().setAccessRestriction(sourceType, accessRestriction);
        TypeParameter[] typeParameters = this.referenceContext.typeParameters;
        sourceType.typeVariables = ((typeParameters == null || typeParameters.length == 0) ? Binding.NO_TYPE_VARIABLES : null);
        sourceType.fPackage.addType(sourceType);
        this.checkAndSetModifiers();
        this.buildTypeVariables();
        this.buildMemberTypes(accessRestriction);
        return sourceType;
    }
    
    private void buildTypeVariables() {
        SourceTypeBinding sourceType = this.referenceContext.binding;
        TypeParameter[] typeParameters = this.referenceContext.typeParameters;
        if (typeParameters == null || typeParameters.length == 0) {
            sourceType.setTypeVariables(Binding.NO_TYPE_VARIABLES);
            return;
        }
        sourceType.setTypeVariables(Binding.NO_TYPE_VARIABLES);
        if (sourceType.id == 1) {
            this.problemReporter().objectCannotBeGeneric(this.referenceContext);
            return;
        }
        sourceType.setTypeVariables(this.createTypeVariables(typeParameters, sourceType));
        sourceType.modifiers |= 0x40000000;
        return;
    }
    
    @Override
    void resolveTypeParameter(TypeParameter typeParameter) {
        typeParameter.resolve(this);
    }
    
    private void checkAndSetModifiers() {
        SourceTypeBinding sourceType = this.referenceContext.binding;
        int modifiers = sourceType.modifiers;
        if ((modifiers & 0x400000) != 0x0) {
            this.problemReporter().duplicateModifierForType(sourceType);
        }
        ReferenceBinding enclosingType = sourceType.enclosingType();
        boolean isMemberType = sourceType.isMemberType();
        if (isMemberType) {
            modifiers |= (enclosingType.modifiers & 0x40000800);
            if (enclosingType.isInterface()) {
                modifiers |= 0x1;
            }
            if (sourceType.isEnum()) {
                if (!enclosingType.isStatic()) {
                    this.problemReporter().nonStaticContextForEnumMemberType(sourceType);
                }
                else {
                    modifiers |= 0x8;
                }
            }
            else if (sourceType.isInterface()) {
                modifiers |= 0x8;
            }
        }
        else if (sourceType.isLocalType()) {
            if (sourceType.isEnum()) {
                this.problemReporter().illegalLocalTypeDeclaration(this.referenceContext);
                sourceType.modifiers = 0;
                return;
            }
            if (sourceType.isAnonymousType()) {
                modifiers |= 0x10;
                if (this.referenceContext.allocation.type == null) {
                    modifiers |= 0x4000;
                }
            }
            Scope scope = this;
            do {
                switch (scope.kind) {
                    case 2: {
                        MethodScope methodScope = (MethodScope)scope;
                        if (methodScope.isLambdaScope()) {
                            methodScope = methodScope.namedMethodScope();
                        }
                        if (methodScope.isInsideInitializer()) {
                            SourceTypeBinding type = ((TypeDeclaration)methodScope.referenceContext).binding;
                            if (methodScope.initializedField != null) {
                                if (methodScope.initializedField.isViewedAsDeprecated() && !sourceType.isDeprecated()) {
                                    modifiers |= 0x200000;
                                    break;
                                }
                                break;
                            }
                            else {
                                if (type.isStrictfp()) {
                                    modifiers |= 0x800;
                                }
                                if (type.isViewedAsDeprecated() && !sourceType.isDeprecated()) {
                                    modifiers |= 0x200000;
                                    break;
                                }
                                break;
                            }
                        }
                        else {
                            MethodBinding method = ((AbstractMethodDeclaration)methodScope.referenceContext).binding;
                            if (method != null) {
                                if (method.isStrictfp()) {
                                    modifiers |= 0x800;
                                }
                                if (method.isViewedAsDeprecated() && !sourceType.isDeprecated()) {
                                    modifiers |= 0x200000;
                                    break;
                                }
                                break;
                            }
                            else {
                                break;
                            }
                        }
                    }
                    case 3: {
                        if (enclosingType.isStrictfp()) {
                            modifiers |= 0x800;
                        }
                        if (enclosingType.isViewedAsDeprecated() && !sourceType.isDeprecated()) {
                            modifiers |= 0x200000;
                            break;
                        }
                        break;
                    }
                }
                scope = scope.parent;
            } while (scope != null);
        }
        int realModifiers = modifiers & 0xFFFF;
        Label_1000: {
            if ((realModifiers & 0x200) != 0x0) {
                if (isMemberType) {
                    if ((realModifiers & 0xFFFFD1F0) != 0x0) {
                        if ((realModifiers & 0x2000) != 0x0) {
                            this.problemReporter().illegalModifierForAnnotationMemberType(sourceType);
                        }
                        else {
                            this.problemReporter().illegalModifierForMemberInterface(sourceType);
                        }
                    }
                }
                else if ((realModifiers & 0xFFFFD1FE) != 0x0) {
                    if ((realModifiers & 0x2000) != 0x0) {
                        this.problemReporter().illegalModifierForAnnotationType(sourceType);
                    }
                    else {
                        this.problemReporter().illegalModifierForInterface(sourceType);
                    }
                }
                if (sourceType.sourceName == TypeConstants.PACKAGE_INFO_NAME && this.compilerOptions().targetJDK > 3211264L) {
                    modifiers |= 0x1000;
                }
                modifiers |= 0x400;
            }
            else if ((realModifiers & 0x4000) != 0x0) {
                if (isMemberType) {
                    if ((realModifiers & 0xFFFFB7F0) != 0x0) {
                        this.problemReporter().illegalModifierForMemberEnum(sourceType);
                        modifiers &= 0xFFFFFBFF;
                        realModifiers &= 0xFFFFFBFF;
                    }
                }
                else if (!sourceType.isLocalType() && (realModifiers & 0xFFFFB7FE) != 0x0) {
                    this.problemReporter().illegalModifierForEnum(sourceType);
                }
                if (!sourceType.isAnonymousType()) {
                    Label_0834: {
                        if ((this.referenceContext.bits & 0x800) != 0x0) {
                            modifiers |= 0x400;
                        }
                        else {
                            TypeDeclaration typeDeclaration = this.referenceContext;
                            FieldDeclaration[] fields = typeDeclaration.fields;
                            int fieldsLength = (fields == null) ? 0 : fields.length;
                            if (fieldsLength != 0) {
                                AbstractMethodDeclaration[] methods = typeDeclaration.methods;
                                int methodsLength = (methods == null) ? 0 : methods.length;
                                boolean definesAbstractMethod = typeDeclaration.superInterfaces != null;
                                for (int i = 0; i < methodsLength && !definesAbstractMethod; definesAbstractMethod = methods[i].isAbstract(), ++i) {}
                                if (definesAbstractMethod) {
                                    boolean needAbstractBit = false;
                                    for (FieldDeclaration fieldDecl : fields) {
                                        if (fieldDecl.getKind() == 3) {
                                            if ((fieldDecl.initialization instanceof QualifiedAllocationExpression)) {
                                                needAbstractBit = true;
                                            }
                                            else {
                                                break Label_0834;
                                            }
                                        }
                                    }
                                    if (needAbstractBit) {
                                        modifiers |= 0x400;
                                    }
                                }
                            }
                        }
                    }
                    TypeDeclaration typeDeclaration = this.referenceContext;
                    FieldDeclaration[] fields = typeDeclaration.fields;
                    if (fields != null) {
                        int k = 0;
                        int fieldsLength2 = fields.length;
                        while (k < fieldsLength2) {
                            FieldDeclaration fieldDecl2 = fields[k];
                            if (fieldDecl2.getKind() == 3 && fieldDecl2.initialization instanceof QualifiedAllocationExpression) {
                                break Label_1000;
                            }
                            ++k;
                        }
                    }
                    modifiers |= 0x10;
                }
            }
            else {
                if (isMemberType) {
                    if ((realModifiers & 0xFFFFF3E0) != 0x0) {
                        this.problemReporter().illegalModifierForMemberClass(sourceType);
                    }
                }
                else if (sourceType.isLocalType()) {
                    if ((realModifiers & 0xFFFFF3EF) != 0x0) {
                        this.problemReporter().illegalModifierForLocalClass(sourceType);
                    }
                }
                else if ((realModifiers & 0xFFFFF3EE) != 0x0) {
                    this.problemReporter().illegalModifierForClass(sourceType);
                }
                if ((realModifiers & 0x410) == 0x410) {
                    this.problemReporter().illegalModifierCombinationFinalAbstractForClass(sourceType);
                }
            }
        }
        if (isMemberType) {
            if (enclosingType.isInterface()) {
                if ((realModifiers & 0x6) != 0x0) {
                    this.problemReporter().illegalVisibilityModifierForInterfaceMemberType(sourceType);
                    if ((realModifiers & 0x4) != 0x0) {
                        modifiers &= 0xFFFFFFFB;
                    }
                    if ((realModifiers & 0x2) != 0x0) {
                        modifiers &= 0xFFFFFFFD;
                    }
                }
            }
            else {
                int accessorBits = realModifiers & 0x7;
                if ((accessorBits & accessorBits - 1) > 1) {
                    this.problemReporter().illegalVisibilityModifierCombinationForMemberType(sourceType);
                    if ((accessorBits & 0x1) != 0x0) {
                        if ((accessorBits & 0x4) != 0x0) {
                            modifiers &= 0xFFFFFFFB;
                        }
                        if ((accessorBits & 0x2) != 0x0) {
                            modifiers &= 0xFFFFFFFD;
                        }
                    }
                    else if ((accessorBits & 0x4) != 0x0 && (accessorBits & 0x2) != 0x0) {
                        modifiers &= 0xFFFFFFFD;
                    }
                }
            }
            if ((realModifiers & 0x8) == 0x0) {
                if (enclosingType.isInterface()) {
                    modifiers |= 0x8;
                }
            }
            else if (!enclosingType.isStatic()) {
                this.problemReporter().illegalStaticModifierForMemberType(sourceType);
            }
        }
        sourceType.modifiers = modifiers;
    }
    
    private void checkAndSetModifiersForField(FieldBinding fieldBinding, FieldDeclaration fieldDecl) {
        int modifiers = fieldBinding.modifiers;
        ReferenceBinding declaringClass = fieldBinding.declaringClass;
        if ((modifiers & 0x400000) != 0x0) {
            this.problemReporter().duplicateModifierForField(declaringClass, fieldDecl);
        }
        if (declaringClass.isInterface()) {
            modifiers |= 0x19;
            if ((modifiers & 0xFFFF) != 0x19) {
                if ((declaringClass.modifiers & 0x2000) != 0x0) {
                    this.problemReporter().illegalModifierForAnnotationField(fieldDecl);
                }
                else {
                    this.problemReporter().illegalModifierForInterfaceField(fieldDecl);
                }
            }
            fieldBinding.modifiers = modifiers;
            return;
        }
        else if (fieldDecl.getKind() == 3) {
            if ((modifiers & 0xFFFF) != 0x0) {
                this.problemReporter().illegalModifierForEnumConstant(declaringClass, fieldDecl);
            }
            fieldBinding.modifiers |= 0x8004019;
            return;
        }
        else {
            int realModifiers = modifiers & 0xFFFF;
            if ((realModifiers & 0xFFFFFF20) != 0x0) {
                this.problemReporter().illegalModifierForField(declaringClass, fieldDecl);
                modifiers &= 0xFFFF00DF;
            }
            int accessorBits = realModifiers & 0x7;
            if ((accessorBits & accessorBits - 1) > 1) {
                this.problemReporter().illegalVisibilityModifierCombinationForField(declaringClass, fieldDecl);
                if ((accessorBits & 0x1) != 0x0) {
                    if ((accessorBits & 0x4) != 0x0) {
                        modifiers &= 0xFFFFFFFB;
                    }
                    if ((accessorBits & 0x2) != 0x0) {
                        modifiers &= 0xFFFFFFFD;
                    }
                }
                else if ((accessorBits & 0x4) != 0x0 && (accessorBits & 0x2) != 0x0) {
                    modifiers &= 0xFFFFFFFD;
                }
            }
            if ((realModifiers & 0x50) == 0x50) {
                this.problemReporter().illegalModifierCombinationFinalVolatileForField(declaringClass, fieldDecl);
            }
            if (fieldDecl.initialization == null && (modifiers & 0x10) != 0x0) {
                modifiers |= 0x4000000;
            }
            fieldBinding.modifiers = modifiers;
            return;
        }
    }
    
    public void checkParameterizedSuperTypeCollisions() {
        SourceTypeBinding sourceType = this.referenceContext.binding;
        ReferenceBinding[] interfaces = sourceType.superInterfaces;
        Map<Object, Object> invocations = new HashMap(2);
        ReferenceBinding itsSuperclass = sourceType.isInterface() ? null : sourceType.superclass;
        for (int i = 0, length = interfaces.length; i < length; ++i) {
            ReferenceBinding one = interfaces[i];
            if (one != null && (itsSuperclass == null || !this.hasErasedCandidatesCollisions(itsSuperclass, one, invocations, sourceType, this.referenceContext))) {
                for (ReferenceBinding two : interfaces) {
                    if (two != null && this.hasErasedCandidatesCollisions(one, two, invocations, sourceType, this.referenceContext)) {
                        break;
                    }
                }
            }
        }
        TypeParameter[] typeParameters = this.referenceContext.typeParameters;
    Label_0356:
        for (int k = 0, paramLength = (typeParameters == null) ? 0 : typeParameters.length; k < paramLength; ++k) {
            TypeParameter typeParameter = typeParameters[k];
            TypeVariableBinding typeVariable = typeParameter.binding;
            if (typeVariable != null && typeVariable.isValidBinding()) {
                TypeReference[] boundRefs = typeParameter.bounds;
                if (boundRefs != null) {
                    boolean checkSuperclass = TypeBinding.equalsEquals(typeVariable.firstBound, typeVariable.superclass);
                    for (int l = 0, boundLength = boundRefs.length; l < boundLength; ++l) {
                        TypeReference typeRef = boundRefs[l];
                        TypeBinding superType = typeRef.resolvedType;
                        if (superType != null && superType.isValidBinding()) {
                            if (checkSuperclass && this.hasErasedCandidatesCollisions(superType, typeVariable.superclass, invocations, typeVariable, typeRef)) {
                                break;
                            }
                            int index = typeVariable.superInterfaces.length;
                            while (--index >= 0) {
                                if (this.hasErasedCandidatesCollisions(superType, typeVariable.superInterfaces[index], invocations, typeVariable, typeRef)) {
                                    continue Label_0356;
                                }
                                continue;
                            }
                        }
                    }
                }
            }
        }
        ReferenceBinding[] memberTypes = this.referenceContext.binding.memberTypes;
        if (memberTypes != null && memberTypes != Binding.NO_MEMBER_TYPES) {
            for (int m = 0, size = memberTypes.length; m < size; ++m) {
                ((SourceTypeBinding)memberTypes[m]).scope.checkParameterizedSuperTypeCollisions();
            }
        }
    }
    
    private void checkForInheritedMemberTypes(SourceTypeBinding sourceType) {
        ReferenceBinding currentType = sourceType;
        ReferenceBinding[] interfacesToVisit = null;
        int nextPosition = 0;
        while (!currentType.hasMemberTypes()) {
            ReferenceBinding[] itsInterfaces = currentType.superInterfaces();
            if (itsInterfaces != null && itsInterfaces != Binding.NO_SUPERINTERFACES) {
                if (interfacesToVisit == null) {
                    interfacesToVisit = itsInterfaces;
                    nextPosition = interfacesToVisit.length;
                }
                else {
                    int itsLength = itsInterfaces.length;
                    if (nextPosition + itsLength >= interfacesToVisit.length) {
                        System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new ReferenceBinding[nextPosition + itsLength + 5], 0, nextPosition);
                    }
                    int a = 0;
                Label_0139:
                    while (a < itsLength) {
                        ReferenceBinding next = itsInterfaces[a];
                        int b = 0;
                        Label_0136: {
                            while (b < nextPosition) {
                                if (TypeBinding.equalsEquals(next, interfacesToVisit[b])) {
                                    ++a;
                                    continue Label_0139;
                                }
                                ++b;
                            }
                            interfacesToVisit[nextPosition++] = next;
                            continue;
                        }
                    }
                }
            }
            if ((currentType = currentType.superclass()) != null && (currentType.tagBits & 0x10000L) == 0x0L) {
                continue;
            }
            if (interfacesToVisit != null) {
                boolean needToTag = false;
                for (int i = 0; i < nextPosition; ++i) {
                    ReferenceBinding anInterface = interfacesToVisit[i];
                    if ((anInterface.tagBits & 0x10000L) == 0x0L) {
                        if (anInterface.hasMemberTypes()) {
                            return;
                        }
                        needToTag = true;
                        ReferenceBinding[] itsInterfaces2 = anInterface.superInterfaces();
                        if (itsInterfaces2 != null && itsInterfaces2 != Binding.NO_SUPERINTERFACES) {
                            int itsLength2 = itsInterfaces2.length;
                            if (nextPosition + itsLength2 >= interfacesToVisit.length) {
                                System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new ReferenceBinding[nextPosition + itsLength2 + 5], 0, nextPosition);
                            }
                            int a2 = 0;
                        Label_0324:
                            while (a2 < itsLength2) {
                                ReferenceBinding next2 = itsInterfaces2[a2];
                                int b2 = 0;
                                Label_0321: {
                                    while (b2 < nextPosition) {
                                        if (TypeBinding.equalsEquals(next2, interfacesToVisit[b2])) {
                                            ++a2;
                                            continue Label_0324;
                                        }
                                        ++b2;
                                    }
                                    interfacesToVisit[nextPosition++] = next2;
                                    continue;
                                }
                            }
                        }
                    }
                }
                if (needToTag) {
                    for (ReferenceBinding referenceBinding : interfacesToVisit) {
                        referenceBinding.tagBits |= 0x10000L;
                    }
                }
            }
            currentType = sourceType;
            do {
                currentType.tagBits |= 0x10000L;
            } while ((currentType = currentType.superclass()) != null && (currentType.tagBits & 0x10000L) == 0x0L);
            return;
        }
    }
    
    public void checkParameterizedTypeBounds() {
        for (int i = 0, l = (this.deferredBoundChecks == null) ? 0 : this.deferredBoundChecks.size(); i < l; ++i) {
            Object toCheck = this.deferredBoundChecks.get(i);
            if ((toCheck instanceof TypeReference)) {
                ((TypeReference)toCheck).checkBounds(this);
            }
            else if ((toCheck instanceof Runnable)) {
                ((Runnable)toCheck).run();
            }
        }
        this.deferredBoundChecks = null;
        ReferenceBinding[] memberTypes = this.referenceContext.binding.memberTypes;
        if (memberTypes != null && memberTypes != Binding.NO_MEMBER_TYPES) {
            for (int j = 0, size = memberTypes.length; j < size; ++j) {
                ((SourceTypeBinding)memberTypes[j]).scope.checkParameterizedTypeBounds();
            }
        }
    }
    
    private void connectMemberTypes() {
        SourceTypeBinding sourceType = this.referenceContext.binding;
        ReferenceBinding[] memberTypes = sourceType.memberTypes;
        if (memberTypes != null && memberTypes != Binding.NO_MEMBER_TYPES) {
            for (int i = 0, size = memberTypes.length; i < size; ++i) {
                ((SourceTypeBinding)memberTypes[i]).scope.connectTypeHierarchy();
            }
        }
    }
    
    private boolean connectSuperclass() {
        SourceTypeBinding sourceType = this.referenceContext.binding;
        if (sourceType.id == 1) {
            sourceType.setSuperClass(null);
            sourceType.setSuperInterfaces(Binding.NO_SUPERINTERFACES);
            if (!sourceType.isClass()) {
                this.problemReporter().objectMustBeClass(sourceType);
            }
            if (this.referenceContext.superclass != null || (this.referenceContext.superInterfaces != null && this.referenceContext.superInterfaces.length > 0)) {
                this.problemReporter().objectCannotHaveSuperTypes(sourceType);
            }
            return true;
        }
        else if (this.referenceContext.superclass == null) {
            if (sourceType.isEnum() && this.compilerOptions().sourceLevel >= 3211264L) {
                return this.connectEnumSuperclass();
            }
            sourceType.setSuperClass(this.getJavaLangObject());
            return !this.detectHierarchyCycle(sourceType, sourceType.superclass, null);
        }
        else {
            TypeReference superclassRef = this.referenceContext.superclass;
            ReferenceBinding superclass = this.findSupertype(superclassRef);
            if (superclass != null) {
                if (!superclass.isClass() && (superclass.tagBits & 0x80L) == 0x0L) {
                    this.problemReporter().superclassMustBeAClass(sourceType, superclassRef, superclass);
                }
                else if (superclass.isFinal()) {
                    this.problemReporter().classExtendFinalClass(sourceType, superclassRef, superclass);
                }
                else if ((superclass.tagBits & 0x40000000L) != 0x0L) {
                    this.problemReporter().superTypeCannotUseWildcard(sourceType, superclassRef, superclass);
                }
                else if (superclass.erasure().id == 41) {
                    this.problemReporter().cannotExtendEnum(sourceType, superclassRef, superclass);
                }
                else {
                    if ((superclass.tagBits & 0x20000L) != 0x0L || !superclassRef.resolvedType.isValidBinding()) {
                        sourceType.setSuperClass(superclass);
                        sourceType.tagBits |= 0x20000L;
                        return superclassRef.resolvedType.isValidBinding();
                    }
                    sourceType.setSuperClass(superclass);
                    sourceType.typeBits |= (superclass.typeBits & 0x13);
                    if ((sourceType.typeBits & 0x3) != 0x0) {
                        sourceType.typeBits |= sourceType.applyCloseableClassWhitelists();
                    }
                    return true;
                }
            }
            sourceType.tagBits |= 0x20000L;
            sourceType.setSuperClass(this.getJavaLangObject());
            if ((sourceType.superclass.tagBits & 0x100L) == 0x0L) {
                this.detectHierarchyCycle(sourceType, sourceType.superclass, null);
            }
            return false;
        }
    }
    
    private boolean connectEnumSuperclass() {
        SourceTypeBinding sourceType = this.referenceContext.binding;
        ReferenceBinding rootEnumType = this.getJavaLangEnum();
        if ((rootEnumType.tagBits & 0x80L) != 0x0L) {
            sourceType.tagBits |= 0x20000L;
            sourceType.setSuperClass(rootEnumType);
            return false;
        }
        boolean foundCycle = this.detectHierarchyCycle(sourceType, rootEnumType, null);
        TypeVariableBinding[] refTypeVariables = rootEnumType.typeVariables();
        if (refTypeVariables == Binding.NO_TYPE_VARIABLES) {
            this.problemReporter().nonGenericTypeCannotBeParameterized(0, null, rootEnumType, new TypeBinding[] { sourceType });
            return false;
        }
        if (1 != refTypeVariables.length) {
            this.problemReporter().incorrectArityForParameterizedType(null, rootEnumType, new TypeBinding[] { sourceType });
            return false;
        }
        ParameterizedTypeBinding superType = this.environment().createParameterizedType(rootEnumType, new TypeBinding[] { this.environment().convertToRawType(sourceType, false) }, null);
        sourceType.tagBits |= (superType.tagBits & 0x20000L);
        sourceType.setSuperClass(superType);
        if (!refTypeVariables[0].boundCheck(superType, sourceType, this, null).isOKbyJLS()) {
            this.problemReporter().typeMismatchError(rootEnumType, refTypeVariables[0], sourceType, null);
        }
        return !foundCycle;
    }
    
    private boolean connectSuperInterfaces() {
        SourceTypeBinding sourceType = this.referenceContext.binding;
        sourceType.setSuperInterfaces(Binding.NO_SUPERINTERFACES);
        if (this.referenceContext.superInterfaces == null) {
            if (sourceType.isAnnotationType() && this.compilerOptions().sourceLevel >= 3211264L) {
                ReferenceBinding annotationType = this.getJavaLangAnnotationAnnotation();
                boolean foundCycle = this.detectHierarchyCycle(sourceType, annotationType, null);
                sourceType.setSuperInterfaces(new ReferenceBinding[] { annotationType });
                return !foundCycle;
            }
            return true;
        }
        else {
            if (sourceType.id == 1) {
                return true;
            }
            boolean noProblems = true;
            int length = this.referenceContext.superInterfaces.length;
            ReferenceBinding[] interfaceBindings = new ReferenceBinding[length];
            int count = 0;
        Label_0432:
            for (int i = 0; i < length; ++i) {
                TypeReference superInterfaceRef = this.referenceContext.superInterfaces[i];
                ReferenceBinding superInterface = this.findSupertype(superInterfaceRef);
                if (superInterface == null) {
                    sourceType.tagBits |= 0x20000L;
                    noProblems = false;
                }
                else {
                    int j = 0;
                    while (j < i) {
                        if (TypeBinding.equalsEquals(interfaceBindings[j], superInterface)) {
                            this.problemReporter().duplicateSuperinterface(sourceType, superInterfaceRef, superInterface);
                            sourceType.tagBits |= 0x20000L;
                            noProblems = false;
                            continue Label_0432;
                        }
                        ++j;
                    }
                    if (!superInterface.isInterface() && (superInterface.tagBits & 0x80L) == 0x0L) {
                        this.problemReporter().superinterfaceMustBeAnInterface(sourceType, superInterfaceRef, superInterface);
                        sourceType.tagBits |= 0x20000L;
                        noProblems = false;
                    }
                    else {
                        if (superInterface.isAnnotationType()) {
                            this.problemReporter().annotationTypeUsedAsSuperinterface(sourceType, superInterfaceRef, superInterface);
                        }
                        if ((superInterface.tagBits & 0x40000000L) != 0x0L) {
                            this.problemReporter().superTypeCannotUseWildcard(sourceType, superInterfaceRef, superInterface);
                            sourceType.tagBits |= 0x20000L;
                            noProblems = false;
                        }
                        else {
                            if ((superInterface.tagBits & 0x20000L) != 0x0L || !superInterfaceRef.resolvedType.isValidBinding()) {
                                sourceType.tagBits |= 0x20000L;
                                noProblems &= superInterfaceRef.resolvedType.isValidBinding();
                            }
                            sourceType.typeBits |= (superInterface.typeBits & 0x13);
                            if ((sourceType.typeBits & 0x3) != 0x0) {
                                sourceType.typeBits |= sourceType.applyCloseableInterfaceWhitelists();
                            }
                            interfaceBindings[count++] = superInterface;
                        }
                    }
                }
            }
            if (count > 0) {
                if (count != length) {
                    System.arraycopy(interfaceBindings, 0, interfaceBindings = new ReferenceBinding[count], 0, count);
                }
                sourceType.setSuperInterfaces(interfaceBindings);
            }
            return noProblems;
        }
    }
    
    void connectTypeHierarchy() {
        SourceTypeBinding sourceType = this.referenceContext.binding;
        CompilationUnitScope compilationUnitScope = this.compilationUnitScope();
        boolean wasAlreadyConnecting = compilationUnitScope.connectingHierarchy;
        compilationUnitScope.connectingHierarchy = true;
        try {
            if ((sourceType.tagBits & 0x100L) == 0x0L) {
                sourceType.tagBits |= 0x100L;
                this.environment().typesBeingConnected.add(sourceType);
                boolean noProblems = this.connectSuperclass();
                noProblems &= this.connectSuperInterfaces();
                this.environment().typesBeingConnected.remove(sourceType);
                sourceType.tagBits |= 0x200L;
                noProblems &= this.connectTypeVariables(this.referenceContext.typeParameters, false);
                sourceType.tagBits |= 0x40000L;
                if (noProblems && sourceType.isHierarchyInconsistent()) {
                    this.problemReporter().hierarchyHasProblems(sourceType);
                }
            }
            this.connectMemberTypes();
        }
        finally {
            compilationUnitScope.connectingHierarchy = wasAlreadyConnecting;
        }
        LookupEnvironment env = this.environment();
        try {
            env.missingClassFileLocation = this.referenceContext;
            this.checkForInheritedMemberTypes(sourceType);
        }
        catch (AbortCompilation e) {
            e.updateContext(this.referenceContext, this.referenceCompilationUnit().compilationResult);
            throw e;
        }
        finally {
            env.missingClassFileLocation = null;
        }
    }
    
    @Override
    public boolean deferCheck(Runnable check) {
        if (this.compilationUnitScope().connectingHierarchy) {
            if (this.deferredBoundChecks == null) {
                this.deferredBoundChecks = new ArrayList();
            }
            this.deferredBoundChecks.add(check);
            return true;
        }
        else {
            return false;
        }
    }
    
    private void connectTypeHierarchyWithoutMembers() {
        if ((this.parent instanceof CompilationUnitScope)) {
            if (((CompilationUnitScope)this.parent).imports == null) {
                ((CompilationUnitScope)this.parent).checkAndSetImports();
            }
        }
        else if ((this.parent instanceof ClassScope)) {
            ((ClassScope)this.parent).connectTypeHierarchyWithoutMembers();
        }
        SourceTypeBinding sourceType = this.referenceContext.binding;
        if ((sourceType.tagBits & 0x100L) != 0x0L) {
            return;
        }
        CompilationUnitScope compilationUnitScope = this.compilationUnitScope();
        boolean wasAlreadyConnecting = compilationUnitScope.connectingHierarchy;
        compilationUnitScope.connectingHierarchy = true;
        try {
            sourceType.tagBits |= 0x100L;
            this.environment().typesBeingConnected.add(sourceType);
            boolean noProblems = this.connectSuperclass();
            noProblems &= this.connectSuperInterfaces();
            this.environment().typesBeingConnected.remove(sourceType);
            sourceType.tagBits |= 0x200L;
            noProblems &= this.connectTypeVariables(this.referenceContext.typeParameters, false);
            sourceType.tagBits |= 0x40000L;
            if (noProblems && sourceType.isHierarchyInconsistent()) {
                this.problemReporter().hierarchyHasProblems(sourceType);
            }
        }
        finally {
            compilationUnitScope.connectingHierarchy = wasAlreadyConnecting;
        }
        return;
    }
    
    public boolean detectHierarchyCycle(TypeBinding superType, TypeReference reference) {
        if (!(superType instanceof ReferenceBinding)) {
            return false;
        }
        if (reference == this.superTypeReference) {
            if (superType.isTypeVariable()) {
                return false;
            }
            if (superType.isParameterizedType()) {
                superType = ((ParameterizedTypeBinding)superType).genericType();
            }
            this.compilationUnitScope().recordSuperTypeReference(superType);
            return this.detectHierarchyCycle(this.referenceContext.binding, (ReferenceBinding)superType, reference);
        }
        else {
            if ((superType.tagBits & 0x100L) == 0x0L && superType instanceof SourceTypeBinding) {
                ((SourceTypeBinding)superType).scope.connectTypeHierarchyWithoutMembers();
            }
            return false;
        }
    }
    
    private boolean detectHierarchyCycle(SourceTypeBinding sourceType, ReferenceBinding superType, TypeReference reference) {
        if (superType.isRawType()) {
            superType = ((RawTypeBinding)superType).genericType();
        }
        if (TypeBinding.equalsEquals(sourceType, superType)) {
            this.problemReporter().hierarchyCircularity(sourceType, superType, reference);
            sourceType.tagBits |= 0x20000L;
            return true;
        }
        Label_0126: {
            if (superType.isMemberType()) {
                ReferenceBinding current = superType.enclosingType();
                while (!current.isHierarchyBeingActivelyConnected() || !TypeBinding.equalsEquals(current, sourceType)) {
                    if ((current = current.enclosingType()) == null) {
                        break Label_0126;
                    }
                    continue;
                }
                this.problemReporter().hierarchyCircularity(sourceType, current, reference);
                sourceType.tagBits |= 0x20000L;
                current.tagBits |= 0x20000L;
                return true;
            }
        }
        if (superType.isBinaryBinding()) {
            if (superType.problemId() != 1 && (superType.tagBits & 0x20000L) != 0x0L) {
                sourceType.tagBits |= 0x20000L;
                this.problemReporter().hierarchyHasProblems(sourceType);
                return true;
            }
            boolean hasCycle = false;
            ReferenceBinding parentType = superType.superclass();
            if (parentType != null) {
                if (TypeBinding.equalsEquals(sourceType, parentType)) {
                    this.problemReporter().hierarchyCircularity(sourceType, superType, reference);
                    sourceType.tagBits |= 0x20000L;
                    superType.tagBits |= 0x20000L;
                    return true;
                }
                if (parentType.isParameterizedType()) {
                    parentType = ((ParameterizedTypeBinding)parentType).genericType();
                }
                hasCycle |= this.detectHierarchyCycle(sourceType, parentType, reference);
                if ((parentType.tagBits & 0x20000L) != 0x0L) {
                    sourceType.tagBits |= 0x20000L;
                    parentType.tagBits |= 0x20000L;
                }
            }
            ReferenceBinding[] itsInterfaces = superType.superInterfaces();
            if (itsInterfaces != null && itsInterfaces != Binding.NO_SUPERINTERFACES) {
                int i = 0;
                int length = itsInterfaces.length;
                while (i < length) {
                    ReferenceBinding anInterface = itsInterfaces[i];
                    if (TypeBinding.equalsEquals(sourceType, anInterface)) {
                        this.problemReporter().hierarchyCircularity(sourceType, superType, reference);
                        sourceType.tagBits |= 0x20000L;
                        superType.tagBits |= 0x20000L;
                        return true;
                    }
                    if (anInterface.isParameterizedType()) {
                        anInterface = ((ParameterizedTypeBinding)anInterface).genericType();
                    }
                    hasCycle |= this.detectHierarchyCycle(sourceType, anInterface, reference);
                    if ((anInterface.tagBits & 0x20000L) != 0x0L) {
                        sourceType.tagBits |= 0x20000L;
                        superType.tagBits |= 0x20000L;
                    }
                    ++i;
                }
            }
            return hasCycle;
        }
        else {
            if (superType.isHierarchyBeingActivelyConnected()) {
                TypeReference ref = ((SourceTypeBinding)superType).scope.superTypeReference;
                if (ref != null && ref.resolvedType != null && ((ReferenceBinding)ref.resolvedType).isHierarchyBeingActivelyConnected()) {
                    this.problemReporter().hierarchyCircularity(sourceType, superType, reference);
                    sourceType.tagBits |= 0x20000L;
                    superType.tagBits |= 0x20000L;
                    return true;
                }
                if (ref != null && ref.resolvedType == null) {
                    char[] referredName = ref.getLastToken();
                    for (SourceTypeBinding type : this.environment().typesBeingConnected) {
                        if (CharOperation.equals(referredName, type.sourceName())) {
                            this.problemReporter().hierarchyCircularity(sourceType, superType, reference);
                            sourceType.tagBits |= 0x20000L;
                            superType.tagBits |= 0x20000L;
                            return true;
                        }
                        continue;
                    }
                }
            }
            if ((superType.tagBits & 0x100L) == 0x0L) {
                ((SourceTypeBinding)superType).scope.connectTypeHierarchyWithoutMembers();
            }
            if ((superType.tagBits & 0x20000L) != 0x0L) {
                sourceType.tagBits |= 0x20000L;
            }
            return false;
        }
    }
    
    private ReferenceBinding findSupertype(TypeReference typeReference) {
        CompilationUnitScope unitScope = this.compilationUnitScope();
        LookupEnvironment env = unitScope.environment;
        try {
            (env.missingClassFileLocation = typeReference).aboutToResolve(this);
            unitScope.recordQualifiedReference(typeReference.getTypeName());
            this.superTypeReference = typeReference;
            ReferenceBinding superType = (ReferenceBinding)typeReference.resolveSuperType(this);
            return superType;
        }
        catch (AbortCompilation e) {
            SourceTypeBinding sourceType = this.referenceContext.binding;
            if (sourceType.superInterfaces == null) {
                sourceType.setSuperInterfaces(Binding.NO_SUPERINTERFACES);
            }
            e.updateContext(typeReference, this.referenceCompilationUnit().compilationResult);
            throw e;
        }
        finally {
            env.missingClassFileLocation = null;
            this.superTypeReference = null;
        }
    }
    
    @Override
    public ProblemReporter problemReporter() {
        MethodScope outerMethodScope;
        if ((outerMethodScope = this.outerMostMethodScope()) == null) {
            ProblemReporter problemReporter = this.referenceCompilationUnit().problemReporter;
            problemReporter.referenceContext = this.referenceContext;
            return problemReporter;
        }
        return outerMethodScope.problemReporter();
    }
    
    public TypeDeclaration referenceType() {
        return this.referenceContext;
    }
    
    @Override
    public boolean hasDefaultNullnessFor(int location) {
        SourceTypeBinding binding = this.referenceContext.binding;
        if (binding != null) {
            int nullDefault = binding.getNullDefault();
            if (nullDefault != 0) {
                return (nullDefault & location) != 0x0;
            }
        }
        return this.parent.hasDefaultNullnessFor(location);
    }
    
    @Override
    public String toString() {
        if (this.referenceContext != null) {
            return "--- Class Scope ---\n\n" + this.referenceContext.binding.toString();
        }
        return "--- Class Scope ---\n\n Binding not initialized";
    }
}
